from defi_services.abis.token.erc20_abi import ERC20_ABI
from defi_services.constants.token_constant import Token
from defi_services.jobs.queriers.state_querier import StateQuerier


class TokenServices:
    def __init__(self, state_service: StateQuerier, chain_id: str):
        self.chain_id = chain_id
        self.state_service = state_service

    def get_service_info(self):
        info = {
            "token": {
                "chain_id": self.chain_id,
                "type": "token"
            }
        }
        return info

    @staticmethod
    def get_data(wallet: str, token: str, decoded_data: dict, block_number: int = "latest", **kwargs):
        token_prices = kwargs.get("token_prices", {})
        decimals_key = f"decimals_{token}_{block_number}".lower()
        balance_key = f"balanceOf_{wallet}_{token}_{block_number}".lower()
        if balance_key in decoded_data:
            balance = decoded_data.get(balance_key) or 0
            decimals = decoded_data.get(decimals_key) or 18
            return balance * token_prices.get(token, 1) / 10 ** decimals

        return None

    def get_function_info(self, wallet: str, token: str, block_number: int = "latest"):
        result = self.get_function_balance_info(wallet, token, block_number)
        result.update(self.get_decimals_info(token, block_number))
        return result

    def get_function_token_erc_20_info(self, tokens: list, fn_name: str, fn_paras=None, block_number: int = 'latest'):
        result = {}
        for token in tokens:
            erc20_token = token
            if token == Token.native_token:
                erc20_token = Token.wrapped_token.get(self.chain_id)
            key = f"{fn_name}_{erc20_token}_{block_number}".lower()
            result.update({
                key: self.state_service.get_function_info(
                    erc20_token, ERC20_ABI, fn_name, fn_paras, block_number
                )
            })

        return result

    def get_function_balance_info(self, wallet: str, token: str, block_number: int = "latest"):
        key = f"balanceOf_{wallet}_{token}_{block_number}".lower()
        if token == Token.native_token:
            return {key: self.state_service.get_native_token_balance_info(wallet, block_number)}

        return {key: self.state_service.get_function_info(token, ERC20_ABI, "balanceOf", [wallet], block_number)}

    def get_decimals_info(self, token: str, block_number: int = "latest"):
        decimals_token = token
        if token == Token.native_token:
            decimals_token = Token.wrapped_token.get(self.chain_id)
        key = f"decimals_{token}_{block_number}".lower()
        return {key: self.state_service.get_function_info(decimals_token, ERC20_ABI, "decimals", [], block_number)}



SUSHISWAP_MASTER_CHEF_ABI = [{
                                 "inputs": [
                                     {"internalType": "contract SushiToken", "name": "_sushi", "type": "address"},
                                     {"internalType": "address", "name": "_devaddr", "type": "address"},
                                     {"internalType": "uint256", "name": "_sushiPerBlock", "type": "uint256"},
                                     {"internalType": "uint256", "name": "_startBlock", "type": "uint256"},
                                     {"internalType": "uint256", "name": "_bonusEndBlock", "type": "uint256"}],
                                 "stateMutability": "nonpayable", "type": "constructor"}, {
                                 "anonymous": False, "inputs": [
        {"indexed": True, "internalType": "address", "name": "user", "type": "address"},
        {"indexed": True, "internalType": "uint256", "name": "pid", "type": "uint256"},
        {"indexed": False, "internalType": "uint256", "name": "amount", "type": "uint256"}], "name": "Deposit",
                                 "type": "event"}, {
                                 "anonymous": False, "inputs": [
        {"indexed": True, "internalType": "address", "name": "user", "type": "address"},
        {"indexed": True, "internalType": "uint256", "name": "pid", "type": "uint256"},
        {"indexed": False, "internalType": "uint256", "name": "amount", "type": "uint256"}],
                                 "name": "EmergencyWithdraw", "type": "event"}, {
                                 "anonymous": False, "inputs": [
        {"indexed": True, "internalType": "address", "name": "previousOwner", "type": "address"},
        {"indexed": True, "internalType": "address", "name": "newOwner", "type": "address"}],
                                 "name": "OwnershipTransferred", "type": "event"}, {
                                 "anonymous": False, "inputs": [
        {"indexed": True, "internalType": "address", "name": "user", "type": "address"},
        {"indexed": True, "internalType": "uint256", "name": "pid", "type": "uint256"},
        {"indexed": False, "internalType": "uint256", "name": "amount", "type": "uint256"}], "name": "Withdraw",
                                 "type": "event"}, {
                                 "inputs": [], "name": "BONUS_MULTIPLIER",
                                 "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
                                 "stateMutability": "view", "type": "function"}, {
                                 "inputs": [{"internalType": "uint256", "name": "_allocPoint", "type": "uint256"},
                                            {"internalType": "contract IERC20", "name": "_lpToken", "type": "address"},
                                            {"internalType": "bool", "name": "_withUpdate", "type": "bool"}],
                                 "name": "add", "outputs": [], "stateMutability": "nonpayable", "type": "function"}, {
                                 "inputs": [], "name": "bonusEndBlock",
                                 "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
                                 "stateMutability": "view", "type": "function"}, {
                                 "inputs": [{"internalType": "uint256", "name": "_pid", "type": "uint256"},
                                            {"internalType": "uint256", "name": "_amount", "type": "uint256"}],
                                 "name": "deposit", "outputs": [], "stateMutability": "nonpayable", "type": "function"},
                             {
                                 "inputs": [{"internalType": "address", "name": "_devaddr", "type": "address"}],
                                 "name": "dev", "outputs": [], "stateMutability": "nonpayable", "type": "function"}, {
                                 "inputs": [], "name": "devaddr",
                                 "outputs": [{"internalType": "address", "name": "", "type": "address"}],
                                 "stateMutability": "view", "type": "function"}, {
                                 "inputs": [{"internalType": "uint256", "name": "_pid", "type": "uint256"}],
                                 "name": "emergencyWithdraw", "outputs": [], "stateMutability": "nonpayable",
                                 "type": "function"}, {
                                 "inputs": [{"internalType": "uint256", "name": "_from", "type": "uint256"},
                                            {"internalType": "uint256", "name": "_to", "type": "uint256"}],
                                 "name": "getMultiplier",
                                 "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
                                 "stateMutability": "view", "type": "function"}, {
                                 "inputs": [], "name": "massUpdatePools", "outputs": [],
                                 "stateMutability": "nonpayable", "type": "function"}, {
                                 "inputs": [{"internalType": "uint256", "name": "_pid", "type": "uint256"}],
                                 "name": "migrate", "outputs": [], "stateMutability": "nonpayable", "type": "function"},
                             {
                                 "inputs": [], "name": "migrator",
                                 "outputs": [{"internalType": "contract IMigratorChef", "name": "", "type": "address"}],
                                 "stateMutability": "view", "type": "function"}, {
                                 "inputs": [], "name": "owner",
                                 "outputs": [{"internalType": "address", "name": "", "type": "address"}],
                                 "stateMutability": "view", "type": "function"}, {
                                 "inputs": [{"internalType": "uint256", "name": "_pid", "type": "uint256"},
                                            {"internalType": "address", "name": "_user", "type": "address"}],
                                 "name": "pendingSushi",
                                 "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
                                 "stateMutability": "view", "type": "function"}, {
                                 "inputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
                                 "name": "poolInfo",
                                 "outputs": [{"internalType": "contract IERC20", "name": "lpToken", "type": "address"},
                                             {"internalType": "uint256", "name": "allocPoint", "type": "uint256"},
                                             {"internalType": "uint256", "name": "lastRewardBlock", "type": "uint256"},
                                             {
                                                 "internalType": "uint256", "name": "accSushiPerShare",
                                                 "type": "uint256"}], "stateMutability": "view", "type": "function"}, {
                                 "inputs": [], "name": "poolLength",
                                 "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
                                 "stateMutability": "view", "type": "function"}, {
                                 "inputs": [], "name": "renounceOwnership", "outputs": [],
                                 "stateMutability": "nonpayable", "type": "function"}, {
                                 "inputs": [{"internalType": "uint256", "name": "_pid", "type": "uint256"},
                                            {"internalType": "uint256", "name": "_allocPoint", "type": "uint256"},
                                            {"internalType": "bool", "name": "_withUpdate", "type": "bool"}],
                                 "name": "set", "outputs": [], "stateMutability": "nonpayable", "type": "function"}, {
                                 "inputs": [{
                                                "internalType": "contract IMigratorChef", "name": "_migrator",
                                                "type": "address"}], "name": "setMigrator", "outputs": [],
                                 "stateMutability": "nonpayable", "type": "function"}, {
                                 "inputs": [], "name": "startBlock",
                                 "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
                                 "stateMutability": "view", "type": "function"}, {
                                 "inputs": [], "name": "sushi",
                                 "outputs": [{"internalType": "contract SushiToken", "name": "", "type": "address"}],
                                 "stateMutability": "view", "type": "function"}, {
                                 "inputs": [], "name": "sushiPerBlock",
                                 "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
                                 "stateMutability": "view", "type": "function"}, {
                                 "inputs": [], "name": "totalAllocPoint",
                                 "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
                                 "stateMutability": "view", "type": "function"}, {
                                 "inputs": [{"internalType": "address", "name": "newOwner", "type": "address"}],
                                 "name": "transferOwnership", "outputs": [], "stateMutability": "nonpayable",
                                 "type": "function"}, {
                                 "inputs": [{"internalType": "uint256", "name": "_pid", "type": "uint256"}],
                                 "name": "updatePool", "outputs": [], "stateMutability": "nonpayable",
                                 "type": "function"}, {
                                 "inputs": [{"internalType": "uint256", "name": "", "type": "uint256"},
                                            {"internalType": "address", "name": "", "type": "address"}],
                                 "name": "userInfo",
                                 "outputs": [{"internalType": "uint256", "name": "amount", "type": "uint256"},
                                             {"internalType": "uint256", "name": "rewardDebt", "type": "uint256"}],
                                 "stateMutability": "view", "type": "function"}, {
                                 "inputs": [{"internalType": "uint256", "name": "_pid", "type": "uint256"},
                                            {"internalType": "uint256", "name": "_amount", "type": "uint256"}],
                                 "name": "withdraw", "outputs": [], "stateMutability": "nonpayable",
                                 "type": "function"}]



import json
UWU_INCENTIVES_ABI=json.loads('''
[
  {
    "inputs": [
      {
        "internalType": "uint128[]",
        "name": "_startTimeOffset",
        "type": "uint128[]"
      },
      {
        "internalType": "uint128[]",
        "name": "_rewardsPerSecond",
        "type": "uint128[]"
      },
      {
        "internalType": "address",
        "name": "_poolConfigurator",
        "type": "address"
      },
      {
        "internalType": "contract IMultiFeeDistribution",
        "name": "_rewardMinter",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "_maxMintable",
        "type": "uint256"
      },
      {
        "internalType": "contract IChefIncentivesController",
        "name": "_incentivesController",
        "type": "address"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "constructor"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "token",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "user",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "balance",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "totalSupply",
        "type": "uint256"
      }
    ],
    "name": "BalanceUpdated",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "previousOwner",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "newOwner",
        "type": "address"
      }
    ],
    "name": "OwnershipTransferred",
    "type": "event"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_token",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "_allocPoint",
        "type": "uint256"
      }
    ],
    "name": "addPool",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address[]",
        "name": "_tokens",
        "type": "address[]"
      },
      {
        "internalType": "uint256[]",
        "name": "_allocPoints",
        "type": "uint256[]"
      }
    ],
    "name": "batchUpdateAllocPoint",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_user",
        "type": "address"
      },
      {
        "internalType": "address[]",
        "name": "_tokens",
        "type": "address[]"
      }
    ],
    "name": "claim",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "name": "claimReceiver",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_user",
        "type": "address"
      },
      {
        "internalType": "address[]",
        "name": "_tokens",
        "type": "address[]"
      }
    ],
    "name": "claimableReward",
    "outputs": [
      {
        "internalType": "uint256[]",
        "name": "",
        "type": "uint256[]"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "name": "emissionSchedule",
    "outputs": [
      {
        "internalType": "uint128",
        "name": "startTimeOffset",
        "type": "uint128"
      },
      {
        "internalType": "uint128",
        "name": "rewardsPerSecond",
        "type": "uint128"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_user",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "_balance",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "_totalSupply",
        "type": "uint256"
      }
    ],
    "name": "handleAction",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "incentivesController",
    "outputs": [
      {
        "internalType": "contract IChefIncentivesController",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "maxMintableTokens",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "mintedTokens",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "owner",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "poolConfigurator",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "name": "poolInfo",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "totalSupply",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "allocPoint",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "lastRewardTime",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "accRewardPerShare",
        "type": "uint256"
      },
      {
        "internalType": "contract IOnwardIncentivesController",
        "name": "onwardIncentives",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "poolLength",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "name": "registeredTokens",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "renounceOwnership",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "rewardMinter",
    "outputs": [
      {
        "internalType": "contract IMultiFeeDistribution",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "rewardsPerSecond",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_user",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_receiver",
        "type": "address"
      }
    ],
    "name": "setClaimReceiver",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_token",
        "type": "address"
      },
      {
        "internalType": "contract IOnwardIncentivesController",
        "name": "_incentives",
        "type": "address"
      }
    ],
    "name": "setOnwardIncentives",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "setup",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "start",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "startTime",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "totalAllocPoint",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "newOwner",
        "type": "address"
      }
    ],
    "name": "transferOwnership",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "name": "userBaseClaimable",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "name": "userInfo",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "rewardDebt",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  }
]
''')


import base58
from web3 import Web3


def hex_to_base58(hex_string):
    if hex_string[:2] in ["0x", "0X"]:
        hex_string = "41" + hex_string[2:]
    bytes_str = bytes.fromhex(hex_string)
    base58_str = base58.b58encode_check(bytes_str)
    return base58_str.decode("UTF-8")


def base58_to_hex(base58_string):
    asc_string = base58.b58decode_check(base58_string)
    hex_string = asc_string.hex()
    hex_string = f"0x{hex_string[2:]}"
    return hex_string


def convert_address_dict(data):
    if isinstance(data, dict):
        result = {}
        for key, value in data.items():
            if Web3.is_address(key):
                key = hex_to_base58(key)
            if isinstance(value, dict):
                value = convert_address_dict(value)
            if Web3.is_address(value):
                value = hex_to_base58(value)
            result[key] = value
        return result

    if isinstance(data, list):
        result = []
        for item in data:
            result.append(convert_address_dict(item))
        return result

    return data



import json

CHAIN_LINK_ABI = json.loads(
    '[{"inputs":[],"name":"decimals","outputs":[{"internalType":"uint8","name":"","type":"uint8"}],'
    '"stateMutability":"view","type":"function"},{"inputs":[],"name":"description","outputs":[{'
    '"internalType":"string","name":"","type":"string"}],"stateMutability":"view","type":"function"},{"inputs":[{'
    '"internalType":"uint80","name":"_roundId","type":"uint80"}],"name":"getRoundData","outputs":[{'
    '"internalType":"uint80","name":"roundId","type":"uint80"},{"internalType":"int256","name":"answer",'
    '"type":"int256"},{"internalType":"uint256","name":"startedAt","type":"uint256"},{"internalType":"uint256",'
    '"name":"updatedAt","type":"uint256"},{"internalType":"uint80","name":"answeredInRound","type":"uint80"}],'
    '"stateMutability":"view","type":"function"},{"inputs":[],"name":"latestRoundData","outputs":[{'
    '"internalType":"uint80","name":"roundId","type":"uint80"},{"internalType":"int256","name":"answer",'
    '"type":"int256"},{"internalType":"uint256","name":"startedAt","type":"uint256"},{"internalType":"uint256",'
    '"name":"updatedAt","type":"uint256"},{"internalType":"uint80","name":"answeredInRound","type":"uint80"}],'
    '"stateMutability":"view","type":"function"},{"inputs":[],"name":"version","outputs":[{"internalType":"uint256",'
    '"name":"","type":"uint256"}],"stateMutability":"view","type":"function"}]')



PANCAKESWAP_V3_FACTORY_ABI = [{"inputs": [{"internalType": "address", "name": "_poolDeployer", "type": "address"}],
                               "stateMutability": "nonpayable", "type": "constructor"}, {"anonymous": False, "inputs": [
    {"indexed": True, "internalType": "uint24", "name": "fee", "type": "uint24"},
    {"indexed": True, "internalType": "int24", "name": "tickSpacing", "type": "int24"}], "name": "FeeAmountEnabled",
                                                                                         "type": "event"},
                              {"anonymous": False,
                               "inputs": [{"indexed": True, "internalType": "uint24", "name": "fee", "type": "uint24"},
                                          {"indexed": False, "internalType": "bool", "name": "whitelistRequested",
                                           "type": "bool"},
                                          {"indexed": False, "internalType": "bool", "name": "enabled",
                                           "type": "bool"}], "name": "FeeAmountExtraInfoUpdated", "type": "event"},
                              {"anonymous": False, "inputs": [
                                  {"indexed": True, "internalType": "address", "name": "oldOwner", "type": "address"},
                                  {"indexed": True, "internalType": "address", "name": "newOwner", "type": "address"}],
                               "name": "OwnerChanged", "type": "event"}, {"anonymous": False, "inputs": [
        {"indexed": True, "internalType": "address", "name": "token0", "type": "address"},
        {"indexed": True, "internalType": "address", "name": "token1", "type": "address"},
        {"indexed": True, "internalType": "uint24", "name": "fee", "type": "uint24"},
        {"indexed": False, "internalType": "int24", "name": "tickSpacing", "type": "int24"},
        {"indexed": False, "internalType": "address", "name": "pool", "type": "address"}], "name": "PoolCreated",
                                                                          "type": "event"}, {"anonymous": False,
                                                                                             "inputs": [
                                                                                                 {"indexed": True,
                                                                                                  "internalType": "address",
                                                                                                  "name": "lmPoolDeployer",
                                                                                                  "type": "address"}],
                                                                                             "name": "SetLmPoolDeployer",
                                                                                             "type": "event"},
                              {"anonymous": False, "inputs": [
                                  {"indexed": True, "internalType": "address", "name": "user", "type": "address"},
                                  {"indexed": False, "internalType": "bool", "name": "verified", "type": "bool"}],
                               "name": "WhiteListAdded", "type": "event"}, {
                                  "inputs": [{"internalType": "address", "name": "pool", "type": "address"},
                                             {"internalType": "address", "name": "recipient", "type": "address"},
                                             {"internalType": "uint128", "name": "amount0Requested", "type": "uint128"},
                                             {"internalType": "uint128", "name": "amount1Requested",
                                              "type": "uint128"}], "name": "collectProtocol",
                                  "outputs": [{"internalType": "uint128", "name": "amount0", "type": "uint128"},
                                              {"internalType": "uint128", "name": "amount1", "type": "uint128"}],
                                  "stateMutability": "nonpayable", "type": "function"}, {
                                  "inputs": [{"internalType": "address", "name": "tokenA", "type": "address"},
                                             {"internalType": "address", "name": "tokenB", "type": "address"},
                                             {"internalType": "uint24", "name": "fee", "type": "uint24"}],
                                  "name": "createPool",
                                  "outputs": [{"internalType": "address", "name": "pool", "type": "address"}],
                                  "stateMutability": "nonpayable", "type": "function"}, {
                                  "inputs": [{"internalType": "uint24", "name": "fee", "type": "uint24"},
                                             {"internalType": "int24", "name": "tickSpacing", "type": "int24"}],
                                  "name": "enableFeeAmount", "outputs": [], "stateMutability": "nonpayable",
                                  "type": "function"},
                              {"inputs": [{"internalType": "uint24", "name": "", "type": "uint24"}],
                               "name": "feeAmountTickSpacing",
                               "outputs": [{"internalType": "int24", "name": "", "type": "int24"}],
                               "stateMutability": "view", "type": "function"},
                              {"inputs": [{"internalType": "uint24", "name": "", "type": "uint24"}],
                               "name": "feeAmountTickSpacingExtraInfo",
                               "outputs": [{"internalType": "bool", "name": "whitelistRequested", "type": "bool"},
                                           {"internalType": "bool", "name": "enabled", "type": "bool"}],
                               "stateMutability": "view", "type": "function"}, {
                                  "inputs": [{"internalType": "address", "name": "", "type": "address"},
                                             {"internalType": "address", "name": "", "type": "address"},
                                             {"internalType": "uint24", "name": "", "type": "uint24"}],
                                  "name": "getPool",
                                  "outputs": [{"internalType": "address", "name": "", "type": "address"}],
                                  "stateMutability": "view", "type": "function"},
                              {"inputs": [], "name": "lmPoolDeployer",
                               "outputs": [{"internalType": "address", "name": "", "type": "address"}],
                               "stateMutability": "view", "type": "function"}, {"inputs": [], "name": "owner",
                                                                                "outputs": [{"internalType": "address",
                                                                                             "name": "",
                                                                                             "type": "address"}],
                                                                                "stateMutability": "view",
                                                                                "type": "function"},
                              {"inputs": [], "name": "poolDeployer",
                               "outputs": [{"internalType": "address", "name": "", "type": "address"}],
                               "stateMutability": "view", "type": "function"}, {
                                  "inputs": [{"internalType": "uint24", "name": "fee", "type": "uint24"},
                                             {"internalType": "bool", "name": "whitelistRequested", "type": "bool"},
                                             {"internalType": "bool", "name": "enabled", "type": "bool"}],
                                  "name": "setFeeAmountExtraInfo", "outputs": [], "stateMutability": "nonpayable",
                                  "type": "function"}, {
                                  "inputs": [{"internalType": "address", "name": "pool", "type": "address"},
                                             {"internalType": "uint32", "name": "feeProtocol0", "type": "uint32"},
                                             {"internalType": "uint32", "name": "feeProtocol1", "type": "uint32"}],
                                  "name": "setFeeProtocol", "outputs": [], "stateMutability": "nonpayable",
                                  "type": "function"}, {
                                  "inputs": [{"internalType": "address", "name": "pool", "type": "address"},
                                             {"internalType": "address", "name": "lmPool", "type": "address"}],
                                  "name": "setLmPool", "outputs": [], "stateMutability": "nonpayable",
                                  "type": "function"},
                              {"inputs": [{"internalType": "address", "name": "_lmPoolDeployer", "type": "address"}],
                               "name": "setLmPoolDeployer", "outputs": [], "stateMutability": "nonpayable",
                               "type": "function"},
                              {"inputs": [{"internalType": "address", "name": "_owner", "type": "address"}],
                               "name": "setOwner", "outputs": [], "stateMutability": "nonpayable", "type": "function"},
                              {"inputs": [{"internalType": "address", "name": "user", "type": "address"},
                                          {"internalType": "bool", "name": "verified", "type": "bool"}],
                               "name": "setWhiteListAddress", "outputs": [], "stateMutability": "nonpayable",
                               "type": "function"}]



import logging
import time

from web3 import Web3

from defi_services.abis.lending.aave_v2_and_forlks.oracle_abi import ORACLE_ABI
from defi_services.abis.lending.aave_v2_and_forlks.staked_incentives_abi import STAKED_INCENTIVES_ABI
from defi_services.abis.lending.trava.trava_lending_pool_abi import TRAVA_LENDING_POOL_ABI
from defi_services.abis.token.erc20_abi import ERC20_ABI
from defi_services.constants.chain_constant import Chain
from defi_services.constants.db_constant import DBConst
from defi_services.constants.entities.lending_constant import Lending
from defi_services.constants.time_constant import TimeConstants
from defi_services.jobs.queriers.state_querier import StateQuerier
from defi_services.services.lending.lending_info.bsc.trava_bsc import TRAVA_BSC
from defi_services.services.lending.lending_info.ethereum.trava_eth import TRAVA_ETH
from defi_services.services.lending.lending_info.fantom.trava_ftm import TRAVA_FTM
from defi_services.services.protocol_services import ProtocolServices

logger = logging.getLogger("Trava Lending Pool State Service")


class TravaInfo:
    mapping = {
        Chain.ethereum: TRAVA_ETH,
        Chain.bsc: TRAVA_BSC,
        Chain.fantom: TRAVA_FTM
    }


class TravaStateService(ProtocolServices):
    def __init__(self, state_service: StateQuerier, chain_id: str = "0x1"):
        super().__init__()
        self.name = f"{chain_id}_{Lending.trava}"
        self.chain_id = chain_id
        self.pool_info = TravaInfo.mapping.get(chain_id)
        self.lending_abi = TRAVA_LENDING_POOL_ABI
        self.incentive_abi = STAKED_INCENTIVES_ABI
        self.oracle_abi = ORACLE_ABI
        self.state_service = state_service

    # BASIC FUNCTION
    def get_service_info(self):
        info = {
            Lending.trava: {
                "chain_id": self.chain_id,
                "type": "lending",
                "protocol_info": self.pool_info
            }
        }
        return info

    def get_dapp_asset_info(self, block_number: int = 'latest'):
        begin = time.time()
        _w3 = self.state_service.get_w3()
        pool_address = Web3.to_checksum_address(self.pool_info['address'])
        contract = _w3.eth.contract(address=pool_address, abi=self.lending_abi)
        reserves_list = contract.functions.getReservesList().call(block_identifier=block_number)
        reserves_info = {}
        for token in reserves_list:
            value = contract.functions.getReserveData(token).call(block_identifier=block_number)
            key = token.lower()
            reserves_info[key] = {}
            reserves_info[key]["tToken"] = value[6].lower()
            reserves_info[key]["dToken"] = value[7].lower()

            risk_param = bin(value[0][0])[2:]
            reserves_info[key]["loanToValue"] = int(risk_param[-15:], 2) / 10 ** 4
            reserves_info[key]["liquidationThreshold"] = int(risk_param[-31:-16], 2) / 10 ** 4
        logger.info(f"Get reserves information in {time.time() - begin}s")
        return reserves_info

    # CALCULATE APY LENDING POOL
    def get_apy_lending_pool_function_info(
            self,
            reserves_info: dict,
            block_number: int = "latest",
    ):
        rpc_calls = {}
        for token_address, value in reserves_info.items():
            reserve_key = f"getReserveData_{self.name}_{token_address}_{block_number}".lower()
            atoken_total_supply_key = f'totalSupply_{value["tToken"]}_{block_number}'.lower()
            debt_token_total_supply_key = f'totalSupply_{value["dToken"]}_{block_number}'.lower()
            decimals_key = f"decimals_{token_address}_{block_number}".lower()

            rpc_calls[reserve_key] = self.get_function_lending_pool_info("getReserveData", [token_address])
            rpc_calls[atoken_total_supply_key] = self.state_service.get_function_info(
                value["tToken"], ERC20_ABI, "totalSupply", block_number=block_number)
            rpc_calls[debt_token_total_supply_key] = self.state_service.get_function_info(
                value["dToken"], ERC20_ABI, "totalSupply", block_number=block_number)
            rpc_calls[decimals_key] = self.state_service.get_function_info(
                token_address, ERC20_ABI, "decimals", block_number=block_number)

        return rpc_calls

    def get_reserve_tokens_metadata(
            self,
            decoded_data: dict,
            reserves_info: dict,
            block_number: int = "latest"
    ):
        reserve_tokens_info = []
        for token_address, reserve_info in reserves_info.items():
            get_reserve_data_call_id = f'getReserveData_{self.name}_{token_address}_{block_number}'.lower()
            reserve_data = decoded_data.get(get_reserve_data_call_id)

            atoken = reserve_data[6].lower()
            debt_token = reserve_data[7].lower()
            decimals_call_id = f"decimals_{token_address}_{block_number}".lower()
            atoken_total_supply_key = f'totalSupply_{atoken}_{block_number}'.lower()
            debt_token_total_supply_key = f'totalSupply_{debt_token}_{block_number}'.lower()

            reserve_tokens_info.append({
                'underlying': token_address,
                'underlying_decimals': decoded_data.get(decimals_call_id),
                'a_token_supply': decoded_data.get(atoken_total_supply_key),
                'd_token_supply': decoded_data.get(debt_token_total_supply_key),
                'supply_apy': reserve_data[3],
                'borrow_apy': reserve_data[4]
            })

        return reserve_tokens_info

    def calculate_apy_lending_pool_function_call(
            self,
            reserves_info: dict,
            decoded_data: dict,
            token_prices: dict,
            pool_token_price: float,
            pool_decimals: int = 18,
            block_number: int = 'latest',
    ):
        reserve_tokens_info = self.get_reserve_tokens_metadata(decoded_data, reserves_info, block_number)

        data = {}
        for token_info in reserve_tokens_info:
            underlying_token = token_info['underlying']
            data[underlying_token] = self._calculate_interest_rates(token_info)

        return {self.pool_info.get("address"): data}

    @classmethod
    def _calculate_interest_rates(cls, token_info: dict):
        total_supply_t = token_info.get('a_token_supply')
        total_supply_d = token_info.get('d_token_supply')

        total_supply = total_supply_t / 10 ** token_info['underlying_decimals']
        total_borrow = total_supply_d / 10 ** token_info['underlying_decimals']

        supply_apy = float(token_info['supply_apy']) / 10 ** 27
        borrow_apy = float(token_info['borrow_apy']) / 10 ** 27

        return {
            'deposit_apy': supply_apy,
            'borrow_apy': borrow_apy,
            'total_deposit': total_supply,
            'total_borrow': total_borrow
        }

    @staticmethod
    def get_apy_lending_pool_deprecated(
            atokens: dict,
            debt_tokens: dict,
            decimals: dict,
            reserves_info: dict,
            asset_data_tokens: dict,
            total_supply_tokens: dict,
            interest_rate: dict,
            token_prices: dict,
            pool_token_price: float,
            pool_decimals: int = 18,
    ):
        for token_address in reserves_info:
            atoken = atokens.get(token_address)
            debt_token = debt_tokens.get(token_address)
            decimal = decimals.get(token_address)
            total_supply_t = total_supply_tokens.get(atoken)
            total_supply_d = total_supply_tokens.get(debt_token)
            asset_data_t = asset_data_tokens.get(atoken)
            asset_data_d = asset_data_tokens.get(debt_token)
            # update deposit, borrow apy
            total_supply_t = total_supply_t / 10 ** decimal
            total_supply_d = total_supply_d / 10 ** decimal
            eps_t = asset_data_t[0] / 10 ** pool_decimals
            eps_d = asset_data_d[0] / 10 ** pool_decimals
            token_price = token_prices.get(token_address)
            if total_supply_t:
                deposit_apr = eps_t * TimeConstants.A_YEAR * pool_token_price / (
                        total_supply_t * token_price)
            else:
                deposit_apr = 0
            if total_supply_d:
                borrow_apr = eps_d * TimeConstants.A_YEAR * pool_token_price / (
                        total_supply_d * token_price)
            else:
                borrow_apr = 0
            interest_rate[token_address].update({
                "utilization": total_supply_d / total_supply_t,
                DBConst.reward_deposit_apy: deposit_apr,
                DBConst.reward_borrow_apy: borrow_apr})
            # update liquidity
            liquidity_log = {
                'total_borrow': {
                    DBConst.amount: total_supply_d,
                    DBConst.value_in_usd: total_supply_d * token_price},
                'total_deposit': {
                    DBConst.amount: total_supply_t,
                    DBConst.value_in_usd: total_supply_t * token_price}
            }
            interest_rate[token_address].update({DBConst.liquidity_change_logs: liquidity_log})

        return interest_rate

    def calculate_apy_lending_pool_function_call_deprecated(
            self,
            reserves_info: dict,
            decoded_data: dict,
            token_prices: dict,
            pool_token_price: float,
            pool_decimals: int = 18,
            block_number: int = 'latest',
    ):
        reserves_data = {}
        for token in reserves_info:
            get_reserve_data_call_id = f'getReserveData_{self.name}_{token}_{block_number}'.lower()
            reserves_data[token.lower()] = decoded_data.get(get_reserve_data_call_id)

        interest_rate, atokens, debt_tokens, decimals, asset_data_tokens = {}, {}, {}, {}, {}
        total_supply_tokens = {}
        for token_address in reserves_info:
            lower_address = token_address.lower()
            reserve_data = reserves_data[lower_address]
            interest_rate[lower_address] = {
                'deposit_apy': float(reserve_data[3]) / 10 ** 27,
                'borrow_apy': float(reserve_data[4]) / 10 ** 27}
            atoken = reserve_data[6].lower()
            debt_token = reserve_data[7].lower()
            decimals_call_id = f"decimals_{token_address}_{block_number}".lower()
            atoken_assets_key = f"assets_{atoken}_{block_number}".lower()
            debt_token_assets_key = f"assets_{debt_token}_{block_number}".lower()
            atoken_total_supply_key = f'totalSupply_{atoken}_{block_number}'.lower()
            debt_token_total_supply_key = f'totalSupply_{debt_token}_{block_number}'.lower()

            atokens[lower_address] = atoken
            debt_tokens[lower_address] = debt_token
            decimals[lower_address] = decoded_data.get(decimals_call_id)
            asset_data_tokens[atoken] = decoded_data.get(atoken_assets_key)
            asset_data_tokens[debt_token] = decoded_data.get(debt_token_assets_key)
            total_supply_tokens[atoken] = decoded_data.get(atoken_total_supply_key)
            total_supply_tokens[debt_token] = decoded_data.get(debt_token_total_supply_key)

        asset_price_key = f"getAssetsPrices_{self.name}_{block_number}".lower()
        if not token_prices and asset_price_key in decoded_data:
            token_prices = {}
            prices = decoded_data.get(asset_price_key)
            for pos in range(len(reserves_info.keys())):
                token_prices[reserves_info[pos].lower()] = prices[pos] / 10 ** pool_decimals

        data = self.get_apy_lending_pool_deprecated(
            atokens, debt_tokens, decimals, reserves_info, asset_data_tokens, total_supply_tokens, interest_rate,
            token_prices, pool_token_price, pool_decimals
        )

        return data

    # WALLET DEPOSIT BORROW BALANCE
    def get_wallet_deposit_borrow_balance_function_info(
            self,
            wallet: str,
            reserves_info: dict,
            block_number: int = "latest",
            health_factor: bool = False
    ):
        rpc_calls = {}
        for token in reserves_info:
            value = reserves_info[token]
            atoken_balance_of_key = f'balanceOf_{value["tToken"]}_{wallet}_{block_number}'.lower()
            debt_token_balance_of_key = f'balanceOf_{value["dToken"]}_{wallet}_{block_number}'.lower()
            decimals_key = f"decimals_{token}_{block_number}".lower()

            rpc_calls[atoken_balance_of_key] = self.state_service.get_function_info(
                value["tToken"], ERC20_ABI, "balanceOf", [wallet], block_number=block_number)
            rpc_calls[debt_token_balance_of_key] = self.state_service.get_function_info(
                value["dToken"], ERC20_ABI, "balanceOf", [wallet], block_number=block_number)

            rpc_calls[decimals_key] = self.state_service.get_function_info(
                token, ERC20_ABI, "decimals", block_number=block_number)
        if health_factor:
            rpc_calls.update(self.get_health_factor_function_info(wallet, reserves_info, block_number))
        return rpc_calls

    def get_wallet_deposit_borrow_balance(
            self,
            reserves_info: dict,
            token_prices: dict,
            decimals: dict,
            deposit_amount: dict,
            borrow_amount: dict
    ):
        result = {}
        for token, info in reserves_info.items():
            decimals_token = decimals.get(token)
            deposit_amount_wallet = deposit_amount.get(token) / 10 ** decimals_token
            borrow_amount_wallet = borrow_amount.get(token) / 10 ** decimals_token
            result[token] = {
                "borrow_amount": borrow_amount_wallet,
                "deposit_amount": deposit_amount_wallet,
                "is_collateral": True if info.get('liquidationThreshold') > 0 else False
            }
            if token_prices:
                deposit_amount_in_usd = deposit_amount_wallet * token_prices.get(token, 0)
                borrow_amount_in_usd = borrow_amount_wallet * token_prices.get(token, 0)
                result[token].update({
                    "borrow_amount_in_usd": borrow_amount_in_usd,
                    "deposit_amount_in_usd": deposit_amount_in_usd,
                })
        return {self.pool_info.get("address"): result}

    def calculate_wallet_deposit_borrow_balance(
            self,
            wallet,
            reserves_info: dict,
            decoded_data: dict,
            token_prices: dict,
            pool_decimals: int = 18,
            block_number: int = 'latest',
            health_factor: bool = False
    ):
        asset_price_key = f"getAssetsPrices_{self.name}_{block_number}".lower()
        if not token_prices and asset_price_key in decoded_data:
            token_prices = {}
            prices = decoded_data.get(asset_price_key)
            for pos in range(len(reserves_info.keys())):
                token_prices[reserves_info[pos].lower()] = prices[pos] / 10 ** pool_decimals

        decimals, deposit_amount, borrow_amount = {}, {}, {}
        for token in reserves_info:
            value = reserves_info[token]
            get_total_deposit_id = f"balanceOf_{value['tToken']}_{wallet}_{block_number}".lower()
            get_total_borrow_id = f"balanceOf_{value['dToken']}_{wallet}_{block_number}".lower()
            get_decimals_id = f"decimals_{token}_{block_number}".lower()
            deposit_amount[token] = decoded_data.get(get_total_deposit_id)
            borrow_amount[token] = decoded_data.get(get_total_borrow_id)
            decimals[token] = decoded_data.get(get_decimals_id)

        data = self.get_wallet_deposit_borrow_balance(
            reserves_info, token_prices, decimals, deposit_amount,
            borrow_amount
        )
        if health_factor:
            hf = self.calculate_health_factor(
                wallet, reserves_info, decoded_data, token_prices, pool_decimals, block_number)
            data.update(hf)
        return data

    # HEALTH FACTOR
    def get_health_factor_function_info(
            self,
            wallet: str,
            reserves_info: dict = None,
            block_number: int = "latest"
    ):
        rpc_calls = {}
        key = f"getUserAccountData_{self.name}_{wallet}_{block_number}".lower()
        pool_address = self.pool_info.get("address")
        rpc_calls[key] = self.state_service.get_function_info(
            pool_address, self.lending_abi, "getUserAccountData", [wallet], block_number)
        return rpc_calls

    def calculate_health_factor(
            self,
            wallet: str,
            reserves_info,
            decoded_data: dict = None,
            token_prices: dict = None,
            pool_decimals: int = 18,
            block_number: int = "latest"
    ):
        key = f"getUserAccountData_{self.name}_{wallet}_{block_number}".lower()
        data = decoded_data.get(key)
        health_factor = 0
        if data[0] and data[1]:
            health_factor = data[5] / 10 ** 18

        if data[0] and not data[1]:
            health_factor = 100
        return {"health_factor": health_factor}

    # REWARDS BALANCE
    def get_rewards_balance_function_info(
            self,
            wallet,
            reserves_info: dict = None,
            block_number: int = "latest"
    ):
        rpc_calls = {}
        tokens = []
        for token, value in reserves_info.items():
            atoken, debt_token = Web3.to_checksum_address(value['tToken']), Web3.to_checksum_address(value['dToken'])
            tokens += [atoken, debt_token]
        key = f"getRewardsBalance_{self.name}_{wallet}_{block_number}".lower()
        rpc_calls[key] = self.get_function_incentive_info(
            "getRewardsBalance", [tokens, wallet], block_number)

        return rpc_calls

    def calculate_rewards_balance(
            self, wallet: str, reserves_info: dict, decoded_data: dict, block_number: int = "latest"):
        reward_token = self.pool_info['rewardToken']
        key = f"getRewardsBalance_{self.name}_{wallet}_{block_number}".lower()
        rewards = decoded_data.get(key) / 10 ** 18
        result = {
            reward_token: {"amount": rewards}
        }

        return result

    def get_function_lending_pool_info(self, fn_name: str, fn_paras=None, block_number: int = 'latest'):
        return self.state_service.get_function_info(
            self.pool_info['address'], self.lending_abi, fn_name, fn_paras, block_number
        )

    def get_function_incentive_info(self, fn_name: str, fn_paras=None, block_number: int = 'latest'):
        return self.state_service.get_function_info(
            self.pool_info['stakedIncentiveAddress'], self.incentive_abi, fn_name, fn_paras, block_number
        )

    def get_function_oracle_info(self, fn_name: str, fn_paras=None, block_number: int = 'latest'):
        return self.state_service.get_function_info(
            self.pool_info['oracleAddress'], self.oracle_abi, fn_name, fn_paras, block_number
        )



import json

WEPIGGY_LENS_ABI = json.loads('''
[
  {
    "inputs": [
      {
        "internalType": "contract ComptrollerLensInterface",
        "name": "_comptroller",
        "type": "address"
      },
      {
        "internalType": "contract PiggyDistributionInterface",
        "name": "_distribution",
        "type": "address"
      },
      {
        "internalType": "string",
        "name": "_pNativeToken",
        "type": "string"
      },
      {
        "internalType": "string",
        "name": "_nativeToken",
        "type": "string"
      },
      {
        "internalType": "string",
        "name": "_nativeName",
        "type": "string"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "constructor"
  },
  {
    "inputs": [],
    "name": "all",
    "outputs": [
      {
        "components": [
          {
            "internalType": "address",
            "name": "pTokenAddress",
            "type": "address"
          },
          {
            "internalType": "uint256",
            "name": "pTokenDecimals",
            "type": "uint256"
          },
          {
            "internalType": "address",
            "name": "underlyingAddress",
            "type": "address"
          },
          {
            "internalType": "uint256",
            "name": "underlyingDecimals",
            "type": "uint256"
          },
          {
            "internalType": "string",
            "name": "underlyingSymbol",
            "type": "string"
          },
          {
            "internalType": "string",
            "name": "underlyingName",
            "type": "string"
          },
          {
            "internalType": "uint256",
            "name": "exchangeRateCurrent",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "supplyRatePerBlock",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "borrowRatePerBlock",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "reserveFactorMantissa",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "collateralFactorMantissa",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "totalBorrows",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "totalReserves",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "totalSupply",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "totalCash",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "price",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "mintCap",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "borrowCap",
            "type": "uint256"
          },
          {
            "internalType": "bool",
            "name": "isListed",
            "type": "bool"
          },
          {
            "internalType": "uint256",
            "name": "blockNumber",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "accrualBlockNumber",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "borrowIndex",
            "type": "uint256"
          }
        ],
        "internalType": "struct WePiggyLensV2.PTokenMetadata[]",
        "name": "",
        "type": "tuple[]"
      },
      {
        "components": [
          {
            "internalType": "contract PTokenLensInterface",
            "name": "market",
            "type": "address"
          },
          {
            "internalType": "uint256",
            "name": "blocksPerYear",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "multiplierPerBlock",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "baseRatePerBlock",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "jumpMultiplierPerBlock",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "kink",
            "type": "uint256"
          }
        ],
        "internalType": "struct WePiggyLensV2.InterestRateModel[]",
        "name": "",
        "type": "tuple[]"
      },
      {
        "components": [
          {
            "internalType": "contract PTokenLensInterface",
            "name": "market",
            "type": "address"
          },
          {
            "internalType": "uint256",
            "name": "wpcSpeed",
            "type": "uint256"
          }
        ],
        "internalType": "struct WePiggyLensV2.PTokenWpcSpeed[]",
        "name": "",
        "type": "tuple[]"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address payable",
        "name": "account",
        "type": "address"
      }
    ],
    "name": "allForAccount",
    "outputs": [
      {
        "components": [
          {
            "internalType": "contract PTokenLensInterface[]",
            "name": "markets",
            "type": "address[]"
          },
          {
            "internalType": "uint256",
            "name": "liquidity",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "shortfall",
            "type": "uint256"
          }
        ],
        "internalType": "struct WePiggyLensV2.AccountLimits",
        "name": "",
        "type": "tuple"
      },
      {
        "components": [
          {
            "internalType": "address",
            "name": "pToken",
            "type": "address"
          },
          {
            "internalType": "uint256",
            "name": "balance",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "borrowBalance",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "exchangeRateMantissa",
            "type": "uint256"
          }
        ],
        "internalType": "struct WePiggyLensV2.PTokenBalances[]",
        "name": "",
        "type": "tuple[]"
      },
      {
        "components": [
          {
            "internalType": "address",
            "name": "pTokenAddress",
            "type": "address"
          },
          {
            "internalType": "uint256",
            "name": "pTokenDecimals",
            "type": "uint256"
          },
          {
            "internalType": "address",
            "name": "underlyingAddress",
            "type": "address"
          },
          {
            "internalType": "uint256",
            "name": "underlyingDecimals",
            "type": "uint256"
          },
          {
            "internalType": "string",
            "name": "underlyingSymbol",
            "type": "string"
          },
          {
            "internalType": "string",
            "name": "underlyingName",
            "type": "string"
          },
          {
            "internalType": "uint256",
            "name": "exchangeRateCurrent",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "supplyRatePerBlock",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "borrowRatePerBlock",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "reserveFactorMantissa",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "collateralFactorMantissa",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "totalBorrows",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "totalReserves",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "totalSupply",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "totalCash",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "price",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "mintCap",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "borrowCap",
            "type": "uint256"
          },
          {
            "internalType": "bool",
            "name": "isListed",
            "type": "bool"
          },
          {
            "internalType": "uint256",
            "name": "blockNumber",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "accrualBlockNumber",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "borrowIndex",
            "type": "uint256"
          }
        ],
        "internalType": "struct WePiggyLensV2.PTokenMetadata[]",
        "name": "",
        "type": "tuple[]"
      },
      {
        "components": [
          {
            "internalType": "contract PTokenLensInterface",
            "name": "market",
            "type": "address"
          },
          {
            "internalType": "uint256",
            "name": "blocksPerYear",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "multiplierPerBlock",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "baseRatePerBlock",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "jumpMultiplierPerBlock",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "kink",
            "type": "uint256"
          }
        ],
        "internalType": "struct WePiggyLensV2.InterestRateModel[]",
        "name": "",
        "type": "tuple[]"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "comptroller",
    "outputs": [
      {
        "internalType": "contract ComptrollerLensInterface",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "distribution",
    "outputs": [
      {
        "internalType": "contract PiggyDistributionInterface",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address payable",
        "name": "account",
        "type": "address"
      }
    ],
    "name": "getAccountLimits",
    "outputs": [
      {
        "components": [
          {
            "internalType": "contract PTokenLensInterface[]",
            "name": "markets",
            "type": "address[]"
          },
          {
            "internalType": "uint256",
            "name": "liquidity",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "shortfall",
            "type": "uint256"
          }
        ],
        "internalType": "struct WePiggyLensV2.AccountLimits",
        "name": "",
        "type": "tuple"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract PTokenLensInterface",
        "name": "pToken",
        "type": "address"
      }
    ],
    "name": "getInterestRateModel",
    "outputs": [
      {
        "components": [
          {
            "internalType": "contract PTokenLensInterface",
            "name": "market",
            "type": "address"
          },
          {
            "internalType": "uint256",
            "name": "blocksPerYear",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "multiplierPerBlock",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "baseRatePerBlock",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "jumpMultiplierPerBlock",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "kink",
            "type": "uint256"
          }
        ],
        "internalType": "struct WePiggyLensV2.InterestRateModel",
        "name": "",
        "type": "tuple"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract PTokenLensInterface[]",
        "name": "pTokens",
        "type": "address[]"
      }
    ],
    "name": "getInterestRateModels",
    "outputs": [
      {
        "components": [
          {
            "internalType": "contract PTokenLensInterface",
            "name": "market",
            "type": "address"
          },
          {
            "internalType": "uint256",
            "name": "blocksPerYear",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "multiplierPerBlock",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "baseRatePerBlock",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "jumpMultiplierPerBlock",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "kink",
            "type": "uint256"
          }
        ],
        "internalType": "struct WePiggyLensV2.InterestRateModel[]",
        "name": "",
        "type": "tuple[]"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract PTokenLensInterface",
        "name": "pToken",
        "type": "address"
      }
    ],
    "name": "getWpcSpeed",
    "outputs": [
      {
        "components": [
          {
            "internalType": "contract PTokenLensInterface",
            "name": "market",
            "type": "address"
          },
          {
            "internalType": "uint256",
            "name": "wpcSpeed",
            "type": "uint256"
          }
        ],
        "internalType": "struct WePiggyLensV2.PTokenWpcSpeed",
        "name": "",
        "type": "tuple"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract PTokenLensInterface[]",
        "name": "pTokens",
        "type": "address[]"
      }
    ],
    "name": "getWpcSpeeds",
    "outputs": [
      {
        "components": [
          {
            "internalType": "contract PTokenLensInterface",
            "name": "market",
            "type": "address"
          },
          {
            "internalType": "uint256",
            "name": "wpcSpeed",
            "type": "uint256"
          }
        ],
        "internalType": "struct WePiggyLensV2.PTokenWpcSpeed[]",
        "name": "",
        "type": "tuple[]"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "nativeName",
    "outputs": [
      {
        "internalType": "string",
        "name": "",
        "type": "string"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "nativeToken",
    "outputs": [
      {
        "internalType": "string",
        "name": "",
        "type": "string"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "owner",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "pNativeToken",
    "outputs": [
      {
        "internalType": "string",
        "name": "",
        "type": "string"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract PTokenLensInterface",
        "name": "pToken",
        "type": "address"
      },
      {
        "internalType": "address payable",
        "name": "account",
        "type": "address"
      }
    ],
    "name": "pTokenBalances",
    "outputs": [
      {
        "components": [
          {
            "internalType": "address",
            "name": "pToken",
            "type": "address"
          },
          {
            "internalType": "uint256",
            "name": "balance",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "borrowBalance",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "exchangeRateMantissa",
            "type": "uint256"
          }
        ],
        "internalType": "struct WePiggyLensV2.PTokenBalances",
        "name": "",
        "type": "tuple"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract PTokenLensInterface[]",
        "name": "pTokens",
        "type": "address[]"
      },
      {
        "internalType": "address payable",
        "name": "account",
        "type": "address"
      }
    ],
    "name": "pTokenBalancesAll",
    "outputs": [
      {
        "components": [
          {
            "internalType": "address",
            "name": "pToken",
            "type": "address"
          },
          {
            "internalType": "uint256",
            "name": "balance",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "borrowBalance",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "exchangeRateMantissa",
            "type": "uint256"
          }
        ],
        "internalType": "struct WePiggyLensV2.PTokenBalances[]",
        "name": "",
        "type": "tuple[]"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract PTokenLensInterface",
        "name": "pToken",
        "type": "address"
      }
    ],
    "name": "pTokenMetadata",
    "outputs": [
      {
        "components": [
          {
            "internalType": "address",
            "name": "pTokenAddress",
            "type": "address"
          },
          {
            "internalType": "uint256",
            "name": "pTokenDecimals",
            "type": "uint256"
          },
          {
            "internalType": "address",
            "name": "underlyingAddress",
            "type": "address"
          },
          {
            "internalType": "uint256",
            "name": "underlyingDecimals",
            "type": "uint256"
          },
          {
            "internalType": "string",
            "name": "underlyingSymbol",
            "type": "string"
          },
          {
            "internalType": "string",
            "name": "underlyingName",
            "type": "string"
          },
          {
            "internalType": "uint256",
            "name": "exchangeRateCurrent",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "supplyRatePerBlock",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "borrowRatePerBlock",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "reserveFactorMantissa",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "collateralFactorMantissa",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "totalBorrows",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "totalReserves",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "totalSupply",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "totalCash",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "price",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "mintCap",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "borrowCap",
            "type": "uint256"
          },
          {
            "internalType": "bool",
            "name": "isListed",
            "type": "bool"
          },
          {
            "internalType": "uint256",
            "name": "blockNumber",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "accrualBlockNumber",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "borrowIndex",
            "type": "uint256"
          }
        ],
        "internalType": "struct WePiggyLensV2.PTokenMetadata",
        "name": "",
        "type": "tuple"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract PTokenLensInterface[]",
        "name": "pTokens",
        "type": "address[]"
      }
    ],
    "name": "pTokenMetadataAll",
    "outputs": [
      {
        "components": [
          {
            "internalType": "address",
            "name": "pTokenAddress",
            "type": "address"
          },
          {
            "internalType": "uint256",
            "name": "pTokenDecimals",
            "type": "uint256"
          },
          {
            "internalType": "address",
            "name": "underlyingAddress",
            "type": "address"
          },
          {
            "internalType": "uint256",
            "name": "underlyingDecimals",
            "type": "uint256"
          },
          {
            "internalType": "string",
            "name": "underlyingSymbol",
            "type": "string"
          },
          {
            "internalType": "string",
            "name": "underlyingName",
            "type": "string"
          },
          {
            "internalType": "uint256",
            "name": "exchangeRateCurrent",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "supplyRatePerBlock",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "borrowRatePerBlock",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "reserveFactorMantissa",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "collateralFactorMantissa",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "totalBorrows",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "totalReserves",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "totalSupply",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "totalCash",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "price",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "mintCap",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "borrowCap",
            "type": "uint256"
          },
          {
            "internalType": "bool",
            "name": "isListed",
            "type": "bool"
          },
          {
            "internalType": "uint256",
            "name": "blockNumber",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "accrualBlockNumber",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "borrowIndex",
            "type": "uint256"
          }
        ],
        "internalType": "struct WePiggyLensV2.PTokenMetadata[]",
        "name": "",
        "type": "tuple[]"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ComptrollerLensInterface",
        "name": "_comptroller",
        "type": "address"
      },
      {
        "internalType": "contract PiggyDistributionInterface",
        "name": "_distribution",
        "type": "address"
      },
      {
        "internalType": "string",
        "name": "_pNativeToken",
        "type": "string"
      },
      {
        "internalType": "string",
        "name": "_nativeToken",
        "type": "string"
      },
      {
        "internalType": "string",
        "name": "_nativeName",
        "type": "string"
      }
    ],
    "name": "updateProperties",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  }
]
''')


import json
from urllib.parse import quote
import requests


class CosmosTokenServices:
    def __init__(self, lcd: str, denom: str):
        self.lcd = lcd
        self.denom = denom

    # queries the balance of all coins for a single account.
    def query_balances(self, address: str):
        responses = []
        endpoint = '/cosmos/bank/v1beta1/balances/'
        try:
            results = json.loads(requests.get(self.lcd + endpoint + address, timeout=60).content)
            responses += results['balances']
            pagination = results['pagination']['next_key']
            while pagination is not None:
                results = json.loads(
                    requests.get(self.lcd + endpoint + '?pagination.key=' + quote(str(pagination)), timeout=60).content)
                responses += results['balances']
                pagination = results['pagination']['next_key']
            return responses
        except Exception:
            raise Exception

    # queries the balance of a given denom for a single account.
    def query_balances_by_denom(self, address, denom):
        endpoint = '/cosmos/bank/v1beta1/balances/'
        try:
            return json.loads(
                requests.get(self.lcd + endpoint + address + '/by_denom?denom=' + denom, timeout=60).content)
        except Exception:
            raise Exception

    def query_token_decimal(self):
        endpoint = '/cosmos/bank/v1beta1/denoms_metadata'
        try:
            return json.loads(requests.get(self.lcd + endpoint, timeout=60).content)
        except Exception:
            raise Exception



import logging

from defi_services.abis.dex.pancakeswap.pancakeswap_lp_token_abi import LP_TOKEN_ABI
from defi_services.abis.token.erc20_abi import ERC20_ABI
from defi_services.constants.chain_constant import Chain
from defi_services.constants.entities.dex_constant import Dex
from defi_services.jobs.queriers.state_querier import StateQuerier
from defi_services.services.dex.dex_info.pancakeswap_info import PANCAKESWAP_V2_BSC_INFO
from defi_services.services.dex.uniswap_v2_service import UniswapV2Services

logger = logging.getLogger("PancakeSwap V2 State Service")


class PancakeSwapV2Info:
    mapping = {
        Chain.bsc: PANCAKESWAP_V2_BSC_INFO
    }


class PancakeSwapV2Services(UniswapV2Services):
    def __init__(self, state_service: StateQuerier, chain_id: str = '0x38'):
        super().__init__(state_service=state_service, chain_id=chain_id)

        self.pool_info = PancakeSwapV2Info.mapping.get(chain_id)
        if self.pool_info:
            self.masterchef_abi = self.pool_info['master_chef_abi']
            self.factory_abi = self.pool_info['factory_abi']

    def get_service_info(self):
        info = {
            Dex.pancake_v2: {
                "chain_id": self.chain_id,
                "type": "dex",
                "protocol_info": self.pool_info
            }
        }
        return info

    # Get all lp tokens
    def get_farming_supported_lp_token(self, limit: int = 1):
        web3 = self.state_service.get_w3()
        masterchef_addr = self.pool_info.get('master_chef_address')

        master_chef_contract = web3.eth.contract(abi=self.masterchef_abi,
                                                 address=web3.to_checksum_address(masterchef_addr))
        pool_length = master_chef_contract.functions.poolLength().call()

        rpc_calls = {}
        for pid in range(0, min(pool_length, limit)):
            query_id = f'lpToken_{masterchef_addr}_{pid}_latest'.lower()
            rpc_calls[query_id] = self.get_masterchef_function_info(fn_name="lpToken", fn_paras=[pid])

        return rpc_calls

    def decode_farming_supported_lp_token(self, decoded_data, ):
        result = {}
        for query_id, value in decoded_data.items():
            # Format query_id: f'lpToken_{self.masterchef_addr}_{pid}_latest'

            pid = int(query_id.split("_")[-2])
            result[value.lower()] = {"farming_pid": pid}

        return result

    # Get lp token info
    def get_lp_token_function_info(self, supplied_data, block_number: int = "latest"):
        rpc_calls = super().get_lp_token_function_info(supplied_data=supplied_data, block_number=block_number)

        masterchef_addr = self.pool_info.get('master_chef_address')

        lp_token_info = supplied_data['lp_token_info']
        for lp_token, info in lp_token_info.items():
            query_id = f'balanceOf_{lp_token}_{masterchef_addr}_{block_number}'.lower()
            rpc_calls[query_id] = self.state_service.get_function_info(
                address=lp_token, abi=LP_TOKEN_ABI, fn_name="balanceOf", fn_paras=[masterchef_addr],
                block_number=block_number)

        return rpc_calls

    def decode_lp_token_info(self, supplied_data, decoded_data, block_number: int = "latest"):
        result = super().decode_lp_token_info(
            supplied_data=supplied_data, decoded_data=decoded_data, block_number=block_number)

        masterchef_addr = self.pool_info.get('master_chef_address')
        lp_token_info = supplied_data['lp_token_info']
        for lp_token, info in lp_token_info.items():
            staked_balance = decoded_data.get(f'balanceOf_{lp_token}_{masterchef_addr}_{block_number}'.lower())
            masterchef_balance = staked_balance / 10 ** lp_token_info.get('decimals', 18) if staked_balance else 0
            result[lp_token].update({"stake_balance": masterchef_balance})

        return result

    # Get balance of token
    def get_balance_of_token_function_info(self, supplied_data, block_number: int = "latest"):
        rpc_calls = super().get_balance_of_token_function_info(
            supplied_data=supplied_data, block_number=block_number
        )

        masterchef_addr = self.pool_info.get('master_chef_address')

        lp_token_info = supplied_data['lp_token_info']
        for lp_token, info in lp_token_info.items():
            query_id = f'balanceOf_{lp_token}_{masterchef_addr}_{block_number}'.lower()
            rpc_calls[query_id] = self.state_service.get_function_info(
                address=lp_token, abi=LP_TOKEN_ABI, fn_name="balanceOf", fn_paras=[masterchef_addr],
                block_number=block_number)

        return rpc_calls

    def decode_balance_of_token_function_info(self, supplied_data, decoded_data, block_number: int = "latest"):
        result = super().decode_balance_of_token_function_info(
            supplied_data=supplied_data, decoded_data=decoded_data, block_number=block_number
        )

        masterchef_addr = self.pool_info.get('master_chef_address')
        lp_token_info = supplied_data['lp_token_info']
        for lp_token, info in lp_token_info.items():
            staked_balance = decoded_data.get(f'balanceOf_{lp_token}_{masterchef_addr}_{block_number}'.lower())
            staked_balance_decimals = staked_balance / 10 ** lp_token_info.get('decimals', 18) if staked_balance else 0
            result[lp_token].update({"stake_balance": staked_balance_decimals})

            for token_key in ["token0", "token1"]:
                token_amount = result[lp_token].get(f'{token_key}_amount', 0)
                total_supply = lp_token_info.get(lp_token, {}).get('total_supply')
                token_stake_amount = token_amount * staked_balance_decimals / total_supply if total_supply else 0
                result[lp_token][f'{token_key}_stake_amount'] = token_stake_amount

        return result

    # User Information
    def get_user_info_function(
            self, wallet: str, supplied_data, stake: bool = True, block_number: int = "latest"):
        rpc_calls = super().get_user_info_function(
            wallet=wallet, supplied_data=supplied_data, stake=stake, block_number=block_number)

        masterchef_addr = self.pool_info.get('master_chef_address')

        lp_token_info = supplied_data['lp_token_info']
        for lp_token, info in lp_token_info.items():
            if stake and (info.get('farming_pid') is not None):
                pid = int(info.get('farming_pid'))

                query_id = f'userInfo_{masterchef_addr}_{[pid, wallet]}_{block_number}'.lower()
                rpc_calls[query_id] = self.get_masterchef_function_info(
                    fn_name="userInfo", fn_paras=[pid, wallet], block_number=block_number)

        return rpc_calls

    def decode_user_info_function(
            self, wallet: str, supplied_data, decoded_data: dict, stake: bool = True, block_number: int = "latest"):

        result = super().decode_user_info_function(
            wallet=wallet, supplied_data=supplied_data, decoded_data=decoded_data,
            stake=stake, block_number=block_number
        )

        masterchef_addr = self.pool_info.get('master_chef_address')

        lp_token_info = supplied_data['lp_token_info']
        for lp_token, info in lp_token_info.items():
            if stake and (info.get('farming_pid') is not None):
                pid = int(info.get('farming_pid'))
                query_id = f'userInfo_{masterchef_addr}_{[pid, wallet]}_{block_number}'.lower()

                user_info = decoded_data.get(query_id)
                stake_amount = user_info[0] / 10 ** info.get('decimals', 18) if user_info else 0

                total_supply = info.get('total_supply')

                result[lp_token]['farming_pid'] = pid
                result[lp_token]['stake_amount'] = stake_amount
                result[lp_token]['tokens'][info['token0']]['stake_amount'] = stake_amount * info.get('token0_amount',
                                                                                                     0) / total_supply if total_supply else 0
                result[lp_token]['tokens'][info['token1']]['stake_amount'] = stake_amount * info.get('token1_amount',
                                                                                                     0) / total_supply if total_supply else 0

        return result

    def update_stake_token_amount_of_wallet(self, user_info, supplied_data: dict = None):
        """Deprecated"""
        user_info_token_amount = {}
        lp_token_info = supplied_data['lp_token_info']
        for lp_token, value in user_info.items():
            pair_info = lp_token_info.get(lp_token)
            amount = value.get("amount", 0)
            user_info_token_amount[lp_token] = {
                "amount": amount
            }
            if value.get("userInfo") and pair_info:
                pid = pair_info.get("pid")
                if not pid:
                    continue
                decimals = value.get("decimals")
                stake_amount = value.get("userInfo", [0])[0] / 10 ** decimals
                if stake_amount > 0:
                    user_info_token_amount[lp_token]["stakeAmount"] = stake_amount
                    token_pair_amount = self.cal_token_amount_lp_token(
                        amount, stake_amount, pair_info)
                    user_info_token_amount[lp_token].update(token_pair_amount)

        return user_info_token_amount

    def cal_token_amount_lp_token(self, lp_token_amount, stake_lp_amount, list_token_info):
        """Deprecated"""
        token0, token1 = list_token_info.get("token0"), list_token_info.get("token1")
        result = {}
        if token0 and token1:
            token0_price, token1_price = list_token_info.get('token0Price'), list_token_info.get('token1Price')
            lp_token_price = list_token_info.get("price")
            lp_token_amount_in_usd = lp_token_amount * lp_token_price
            stake_lp_amount_in_usd = stake_lp_amount * lp_token_price
            token0_amount = lp_token_amount_in_usd / 2 / token0_price
            token1_amount = lp_token_amount_in_usd / 2 / token1_price
            result = {
                "valueInUSD": lp_token_amount_in_usd,
                "stakeValueInUSD": stake_lp_amount_in_usd,
                token0: {
                    "amount": token0_amount,
                    "stakeAmount": stake_lp_amount_in_usd / 2 / token0_price,
                    "valueInUSD": lp_token_amount_in_usd / 2,
                    "stakeValueInUSD": stake_lp_amount_in_usd / 2,
                },
                token1: {
                    "amount": token1_amount,
                    "stakeAmount": stake_lp_amount_in_usd / 2 / token1_price,
                    "valueInUSD": lp_token_amount_in_usd / 2,
                    "stakeValueInUSD": stake_lp_amount_in_usd / 2,
                }
            }

        return result

    # User Reward
    def get_rewards_balance_function_info(self, wallet, supplied_data, block_number: int = "latest"):
        rpc_calls = {}

        reward_token = self.pool_info.get("reward_token")
        decimals_query_id = f'decimals_{reward_token}_{block_number}'.lower()
        rpc_calls[decimals_query_id] = self.state_service.get_function_info(
            address=reward_token, abi=ERC20_ABI, fn_name="decimals", block_number=block_number)

        masterchef_addr = self.pool_info.get('master_chef_address')

        lp_token_info = supplied_data['lp_token_info']
        for lp_token, info in lp_token_info.items():
            if info.get('farming_pid') is not None:
                pid = int(info.get('farming_pid'))

                query_id = f'pendingCake_{masterchef_addr}_{[pid, wallet]}_{block_number}'.lower()
                rpc_calls[query_id] = self.get_masterchef_function_info(
                    fn_name="pendingCake", fn_paras=[int(pid), wallet], block_number=block_number)

        return rpc_calls

    def calculate_rewards_balance(self, wallet: str, supplied_data: dict, decoded_data: dict,
                                  block_number: int = "latest") -> dict:
        reward_token = self.pool_info.get("reward_token")
        reward_decimals = decoded_data.get(f'decimals_{reward_token}_{block_number}'.lower())

        result = {}

        masterchef_addr = self.pool_info.get('master_chef_address')

        lp_token_info = supplied_data['lp_token_info']
        for lp_token, info in lp_token_info.items():
            if info.get('farming_pid') is not None:
                pid = int(info.get('farming_pid'))
                query_id = f'pendingCake_{masterchef_addr}_{[pid, wallet]}_{block_number}'.lower()

                result[lp_token] = {reward_token: {'amount': decoded_data.get(query_id) / 10 ** reward_decimals}}

        return result

    def get_masterchef_function_info(self, fn_name, fn_paras, block_number: int = 'latest'):
        masterchef_addr = self.pool_info['master_chef_address']
        return self.state_service.get_function_info(
            masterchef_addr, self.masterchef_abi, fn_name, fn_paras, block_number
        )




import json

REP = json.loads("""
[
  {
    "constant": true,
    "inputs": [],
    "name": "getController",
    "outputs": [
      {
        "name": "",
        "type": "address"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": false,
    "inputs": [
      {
        "name": "_controller",
        "type": "address"
      }
    ],
    "name": "setController",
    "outputs": [
      {
        "name": "",
        "type": "bool"
      }
    ],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [],
    "name": "controllerLookupName",
    "outputs": [
      {
        "name": "",
        "type": "bytes32"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "name": "_controller",
        "type": "address"
      },
      {
        "name": "_controllerLookupName",
        "type": "bytes32"
      }
    ],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "constructor"
  },
  {
    "payable": true,
    "stateMutability": "payable",
    "type": "fallback"
  }
]
""")
        


import json

LIDO_BURNER_ABI = json.loads('''
[
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_admin",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_treasury",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_stETH",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "_totalCoverSharesBurnt",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "_totalNonCoverSharesBurnt",
        "type": "uint256"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "constructor"
  },
  {
    "inputs": [],
    "name": "AppAuthLidoFailed",
    "type": "error"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "requestedAmount",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "actualAmount",
        "type": "uint256"
      }
    ],
    "name": "BurnAmountExceedsActual",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "DirectETHTransfer",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "StETHRecoveryWrongFunc",
    "type": "error"
  },
  {
    "inputs": [
      {
        "internalType": "string",
        "name": "field",
        "type": "string"
      }
    ],
    "name": "ZeroAddress",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "ZeroBurnAmount",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "ZeroRecoveryAmount",
    "type": "error"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "requestedBy",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "token",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "ERC20Recovered",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "requestedBy",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "token",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "tokenId",
        "type": "uint256"
      }
    ],
    "name": "ERC721Recovered",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "requestedBy",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "amountOfStETH",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "amountOfShares",
        "type": "uint256"
      }
    ],
    "name": "ExcessStETHRecovered",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "bytes32",
        "name": "role",
        "type": "bytes32"
      },
      {
        "indexed": true,
        "internalType": "bytes32",
        "name": "previousAdminRole",
        "type": "bytes32"
      },
      {
        "indexed": true,
        "internalType": "bytes32",
        "name": "newAdminRole",
        "type": "bytes32"
      }
    ],
    "name": "RoleAdminChanged",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "bytes32",
        "name": "role",
        "type": "bytes32"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "account",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "sender",
        "type": "address"
      }
    ],
    "name": "RoleGranted",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "bytes32",
        "name": "role",
        "type": "bytes32"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "account",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "sender",
        "type": "address"
      }
    ],
    "name": "RoleRevoked",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "bool",
        "name": "isCover",
        "type": "bool"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "requestedBy",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "amountOfStETH",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "amountOfShares",
        "type": "uint256"
      }
    ],
    "name": "StETHBurnRequested",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "bool",
        "name": "isCover",
        "type": "bool"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "amountOfStETH",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "amountOfShares",
        "type": "uint256"
      }
    ],
    "name": "StETHBurnt",
    "type": "event"
  },
  {
    "inputs": [],
    "name": "DEFAULT_ADMIN_ROLE",
    "outputs": [
      {
        "internalType": "bytes32",
        "name": "",
        "type": "bytes32"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "REQUEST_BURN_MY_STETH_ROLE",
    "outputs": [
      {
        "internalType": "bytes32",
        "name": "",
        "type": "bytes32"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "REQUEST_BURN_SHARES_ROLE",
    "outputs": [
      {
        "internalType": "bytes32",
        "name": "",
        "type": "bytes32"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "STETH",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "TREASURY",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "_sharesToBurn",
        "type": "uint256"
      }
    ],
    "name": "commitSharesToBurn",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getCoverSharesBurnt",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getExcessStETH",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getNonCoverSharesBurnt",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "bytes32",
        "name": "role",
        "type": "bytes32"
      }
    ],
    "name": "getRoleAdmin",
    "outputs": [
      {
        "internalType": "bytes32",
        "name": "",
        "type": "bytes32"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "bytes32",
        "name": "role",
        "type": "bytes32"
      },
      {
        "internalType": "uint256",
        "name": "index",
        "type": "uint256"
      }
    ],
    "name": "getRoleMember",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "bytes32",
        "name": "role",
        "type": "bytes32"
      }
    ],
    "name": "getRoleMemberCount",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getSharesRequestedToBurn",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "coverShares",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "nonCoverShares",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "bytes32",
        "name": "role",
        "type": "bytes32"
      },
      {
        "internalType": "address",
        "name": "account",
        "type": "address"
      }
    ],
    "name": "grantRole",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "bytes32",
        "name": "role",
        "type": "bytes32"
      },
      {
        "internalType": "address",
        "name": "account",
        "type": "address"
      }
    ],
    "name": "hasRole",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_token",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "_amount",
        "type": "uint256"
      }
    ],
    "name": "recoverERC20",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_token",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "_tokenId",
        "type": "uint256"
      }
    ],
    "name": "recoverERC721",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "recoverExcessStETH",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "bytes32",
        "name": "role",
        "type": "bytes32"
      },
      {
        "internalType": "address",
        "name": "account",
        "type": "address"
      }
    ],
    "name": "renounceRole",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "_stETHAmountToBurn",
        "type": "uint256"
      }
    ],
    "name": "requestBurnMyStETH",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "_stETHAmountToBurn",
        "type": "uint256"
      }
    ],
    "name": "requestBurnMyStETHForCover",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_from",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "_sharesAmountToBurn",
        "type": "uint256"
      }
    ],
    "name": "requestBurnShares",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_from",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "_sharesAmountToBurn",
        "type": "uint256"
      }
    ],
    "name": "requestBurnSharesForCover",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "bytes32",
        "name": "role",
        "type": "bytes32"
      },
      {
        "internalType": "address",
        "name": "account",
        "type": "address"
      }
    ],
    "name": "revokeRole",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "bytes4",
        "name": "interfaceId",
        "type": "bytes4"
      }
    ],
    "name": "supportsInterface",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "stateMutability": "payable",
    "type": "receive"
  }
]
''')


TRAVA_VAULT_ABI = [
    {
        "inputs": [
            {
                "internalType": "contract IBEP20",
                "name": "stakedToken",
                "type": "address"
            },
            {
                "internalType": "contract IBEP20",
                "name": "rewardToken",
                "type": "address"
            },
            {
                "internalType": "uint256",
                "name": "cooldownSeconds",
                "type": "uint256"
            },
            {
                "internalType": "uint256",
                "name": "unstakeWindow",
                "type": "uint256"
            },
            {
                "internalType": "address",
                "name": "rewardsVault",
                "type": "address"
            },
            {
                "internalType": "address",
                "name": "emissionManager",
                "type": "address"
            },
            {
                "internalType": "uint128",
                "name": "distributionDuration",
                "type": "uint128"
            },
            {
                "internalType": "string",
                "name": "name",
                "type": "string"
            },
            {
                "internalType": "string",
                "name": "symbol",
                "type": "string"
            },
            {
                "internalType": "uint8",
                "name": "decimals",
                "type": "uint8"
            },
            {
                "internalType": "address",
                "name": "governance",
                "type": "address"
            }
        ],
        "stateMutability": "nonpayable",
        "type": "constructor"
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "owner",
                "type": "address"
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "spender",
                "type": "address"
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "value",
                "type": "uint256"
            }
        ],
        "name": "Approval",
        "type": "event"
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "asset",
                "type": "address"
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "emission",
                "type": "uint256"
            }
        ],
        "name": "AssetConfigUpdated",
        "type": "event"
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "asset",
                "type": "address"
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "index",
                "type": "uint256"
            }
        ],
        "name": "AssetIndexUpdated",
        "type": "event"
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "user",
                "type": "address"
            }
        ],
        "name": "Cooldown",
        "type": "event"
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "previousOwner",
                "type": "address"
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "newOwner",
                "type": "address"
            }
        ],
        "name": "OwnershipTransferred",
        "type": "event"
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "from",
                "type": "address"
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "to",
                "type": "address"
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "amount",
                "type": "uint256"
            }
        ],
        "name": "Redeem",
        "type": "event"
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": False,
                "internalType": "address",
                "name": "user",
                "type": "address"
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "amount",
                "type": "uint256"
            }
        ],
        "name": "RewardsAccrued",
        "type": "event"
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "from",
                "type": "address"
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "to",
                "type": "address"
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "amount",
                "type": "uint256"
            }
        ],
        "name": "RewardsClaimed",
        "type": "event"
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "from",
                "type": "address"
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "onBehalfOf",
                "type": "address"
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "amount",
                "type": "uint256"
            }
        ],
        "name": "Staked",
        "type": "event"
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "from",
                "type": "address"
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "to",
                "type": "address"
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "value",
                "type": "uint256"
            }
        ],
        "name": "Transfer",
        "type": "event"
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "user",
                "type": "address"
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "asset",
                "type": "address"
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "index",
                "type": "uint256"
            }
        ],
        "name": "UserIndexUpdated",
        "type": "event"
    },
    {
        "inputs": [],
        "name": "COOLDOWN_SECONDS",
        "outputs": [
            {
                "internalType": "uint256",
                "name": "",
                "type": "uint256"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [],
        "name": "DISTRIBUTION_END",
        "outputs": [
            {
                "internalType": "uint256",
                "name": "",
                "type": "uint256"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [],
        "name": "DOMAIN_SEPARATOR",
        "outputs": [
            {
                "internalType": "bytes32",
                "name": "",
                "type": "bytes32"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [],
        "name": "EIP712_REVISION",
        "outputs": [
            {
                "internalType": "bytes",
                "name": "",
                "type": "bytes"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [],
        "name": "EMISSION_MANAGER",
        "outputs": [
            {
                "internalType": "address",
                "name": "",
                "type": "address"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [],
        "name": "GOVERNANCE",
        "outputs": [
            {
                "internalType": "address",
                "name": "",
                "type": "address"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [],
        "name": "PERMIT_TYPEHASH",
        "outputs": [
            {
                "internalType": "bytes32",
                "name": "",
                "type": "bytes32"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [],
        "name": "PRECISION",
        "outputs": [
            {
                "internalType": "uint8",
                "name": "",
                "type": "uint8"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [],
        "name": "REVISION",
        "outputs": [
            {
                "internalType": "uint256",
                "name": "",
                "type": "uint256"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [],
        "name": "REWARDS_VAULT",
        "outputs": [
            {
                "internalType": "address",
                "name": "",
                "type": "address"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [],
        "name": "REWARD_TOKEN",
        "outputs": [
            {
                "internalType": "contract IBEP20",
                "name": "",
                "type": "address"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [],
        "name": "STAKED_TOKEN",
        "outputs": [
            {
                "internalType": "contract IBEP20",
                "name": "",
                "type": "address"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [],
        "name": "START_TIME_CLAIM",
        "outputs": [
            {
                "internalType": "uint256",
                "name": "",
                "type": "uint256"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [],
        "name": "UNSTAKE_WINDOW",
        "outputs": [
            {
                "internalType": "uint256",
                "name": "",
                "type": "uint256"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [],
        "name": "VESTING_TIME_INTERVAL",
        "outputs": [
            {
                "internalType": "uint256",
                "name": "",
                "type": "uint256"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "",
                "type": "address"
            }
        ],
        "name": "_nonces",
        "outputs": [
            {
                "internalType": "uint256",
                "name": "",
                "type": "uint256"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "owner",
                "type": "address"
            },
            {
                "internalType": "address",
                "name": "spender",
                "type": "address"
            }
        ],
        "name": "allowance",
        "outputs": [
            {
                "internalType": "uint256",
                "name": "",
                "type": "uint256"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "spender",
                "type": "address"
            },
            {
                "internalType": "uint256",
                "name": "amount",
                "type": "uint256"
            }
        ],
        "name": "approve",
        "outputs": [
            {
                "internalType": "bool",
                "name": "",
                "type": "bool"
            }
        ],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "",
                "type": "address"
            }
        ],
        "name": "assets",
        "outputs": [
            {
                "internalType": "uint128",
                "name": "emissionPerSecond",
                "type": "uint128"
            },
            {
                "internalType": "uint128",
                "name": "lastUpdateTimestamp",
                "type": "uint128"
            },
            {
                "internalType": "uint256",
                "name": "index",
                "type": "uint256"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "account",
                "type": "address"
            }
        ],
        "name": "balanceOf",
        "outputs": [
            {
                "internalType": "uint256",
                "name": "",
                "type": "uint256"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "to",
                "type": "address"
            },
            {
                "internalType": "uint256",
                "name": "amount",
                "type": "uint256"
            }
        ],
        "name": "claimRewards",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [
            {
                "components": [
                    {
                        "internalType": "uint104",
                        "name": "emissionPerSecond",
                        "type": "uint104"
                    },
                    {
                        "internalType": "uint256",
                        "name": "totalStaked",
                        "type": "uint256"
                    },
                    {
                        "internalType": "address",
                        "name": "underlyingAsset",
                        "type": "address"
                    }
                ],
                "internalType": "struct DistributionTypes.AssetConfigInput[]",
                "name": "assetsConfigInput",
                "type": "tuple[]"
            }
        ],
        "name": "configureAssets",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [],
        "name": "cooldown",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [],
        "name": "decimals",
        "outputs": [
            {
                "internalType": "uint8",
                "name": "",
                "type": "uint8"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "spender",
                "type": "address"
            },
            {
                "internalType": "uint256",
                "name": "subtractedValue",
                "type": "uint256"
            }
        ],
        "name": "decreaseAllowance",
        "outputs": [
            {
                "internalType": "bool",
                "name": "",
                "type": "bool"
            }
        ],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "lendingPoolAddress",
                "type": "address"
            },
            {
                "internalType": "address",
                "name": "onBehalfOf",
                "type": "address"
            },
            {
                "internalType": "uint256",
                "name": "amount",
                "type": "uint256"
            }
        ],
        "name": "depositToPool",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "staker",
                "type": "address"
            }
        ],
        "name": "getClaimableRewardsBalance",
        "outputs": [
            {
                "internalType": "uint256",
                "name": "",
                "type": "uint256"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "uint256",
                "name": "fromCooldownTimestamp",
                "type": "uint256"
            },
            {
                "internalType": "uint256",
                "name": "amountToReceive",
                "type": "uint256"
            },
            {
                "internalType": "address",
                "name": "toAddress",
                "type": "address"
            },
            {
                "internalType": "uint256",
                "name": "toBalance",
                "type": "uint256"
            }
        ],
        "name": "getNextCooldownTimestamp",
        "outputs": [
            {
                "internalType": "uint256",
                "name": "",
                "type": "uint256"
            }
        ],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [],
        "name": "getOwner",
        "outputs": [
            {
                "internalType": "address",
                "name": "",
                "type": "address"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "staker",
                "type": "address"
            }
        ],
        "name": "getTotalRewardsBalance",
        "outputs": [
            {
                "internalType": "uint256",
                "name": "",
                "type": "uint256"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "user",
                "type": "address"
            },
            {
                "internalType": "address",
                "name": "asset",
                "type": "address"
            }
        ],
        "name": "getUserAssetData",
        "outputs": [
            {
                "internalType": "uint256",
                "name": "",
                "type": "uint256"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "uint256",
                "name": "amount",
                "type": "uint256"
            },
            {
                "internalType": "address",
                "name": "to",
                "type": "address"
            }
        ],
        "name": "harvest",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "spender",
                "type": "address"
            },
            {
                "internalType": "uint256",
                "name": "addedValue",
                "type": "uint256"
            }
        ],
        "name": "increaseAllowance",
        "outputs": [
            {
                "internalType": "bool",
                "name": "",
                "type": "bool"
            }
        ],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "string",
                "name": "name",
                "type": "string"
            },
            {
                "internalType": "string",
                "name": "symbol",
                "type": "string"
            },
            {
                "internalType": "uint8",
                "name": "decimals",
                "type": "uint8"
            }
        ],
        "name": "initialize",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "uint256",
                "name": "amount",
                "type": "uint256"
            }
        ],
        "name": "mint",
        "outputs": [
            {
                "internalType": "bool",
                "name": "",
                "type": "bool"
            }
        ],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [],
        "name": "name",
        "outputs": [
            {
                "internalType": "string",
                "name": "",
                "type": "string"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [],
        "name": "owner",
        "outputs": [
            {
                "internalType": "address",
                "name": "",
                "type": "address"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "owner",
                "type": "address"
            },
            {
                "internalType": "address",
                "name": "spender",
                "type": "address"
            },
            {
                "internalType": "uint256",
                "name": "value",
                "type": "uint256"
            },
            {
                "internalType": "uint256",
                "name": "deadline",
                "type": "uint256"
            },
            {
                "internalType": "uint8",
                "name": "v",
                "type": "uint8"
            },
            {
                "internalType": "bytes32",
                "name": "r",
                "type": "bytes32"
            },
            {
                "internalType": "bytes32",
                "name": "s",
                "type": "bytes32"
            }
        ],
        "name": "permit",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "to",
                "type": "address"
            },
            {
                "internalType": "uint256",
                "name": "amount",
                "type": "uint256"
            }
        ],
        "name": "redeem",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [],
        "name": "renounceOwnership",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "uint256",
                "name": "_startTimeClaim",
                "type": "uint256"
            },
            {
                "internalType": "uint256",
                "name": "_vestingTimeInterval",
                "type": "uint256"
            }
        ],
        "name": "setClaimRewardsTimeline",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "uint256",
                "name": "_coolDownSecond",
                "type": "uint256"
            }
        ],
        "name": "setCoolDownSecond",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "uint256",
                "name": "_distributionEnd",
                "type": "uint256"
            }
        ],
        "name": "setDistributionEnd",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "_emissionManager",
                "type": "address"
            }
        ],
        "name": "setEmissionManager",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "_gov",
                "type": "address"
            }
        ],
        "name": "setGovernance",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "contract IBEP20",
                "name": "_trava",
                "type": "address"
            }
        ],
        "name": "setRewardToken",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "_reward_vault",
                "type": "address"
            }
        ],
        "name": "setRewardVault",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "uint256",
                "name": "_unstake_window",
                "type": "uint256"
            }
        ],
        "name": "setUnstakeWindow",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "onBehalfOf",
                "type": "address"
            },
            {
                "internalType": "uint256",
                "name": "amount",
                "type": "uint256"
            }
        ],
        "name": "stake",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "",
                "type": "address"
            }
        ],
        "name": "stakerRewardsToClaim",
        "outputs": [
            {
                "internalType": "uint256",
                "name": "",
                "type": "uint256"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "",
                "type": "address"
            }
        ],
        "name": "stakersCooldowns",
        "outputs": [
            {
                "internalType": "uint256",
                "name": "",
                "type": "uint256"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [],
        "name": "symbol",
        "outputs": [
            {
                "internalType": "string",
                "name": "",
                "type": "string"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [],
        "name": "totalSupply",
        "outputs": [
            {
                "internalType": "uint256",
                "name": "",
                "type": "uint256"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "recipient",
                "type": "address"
            },
            {
                "internalType": "uint256",
                "name": "amount",
                "type": "uint256"
            }
        ],
        "name": "transfer",
        "outputs": [
            {
                "internalType": "bool",
                "name": "",
                "type": "bool"
            }
        ],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "sender",
                "type": "address"
            },
            {
                "internalType": "address",
                "name": "recipient",
                "type": "address"
            },
            {
                "internalType": "uint256",
                "name": "amount",
                "type": "uint256"
            }
        ],
        "name": "transferFrom",
        "outputs": [
            {
                "internalType": "bool",
                "name": "",
                "type": "bool"
            }
        ],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "newOwner",
                "type": "address"
            }
        ],
        "name": "transferOwnership",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function"
    }
]



import logging

from query_state_lib.base.mappers.eth_call_mapper import EthCall
from query_state_lib.base.mappers.get_balance_mapper import GetBalance
from query_state_lib.base.utils.encoder import encode_eth_call_data
from query_state_lib.client.client_querier import ClientQuerier
from web3 import Web3
from web3.middleware import geth_poa_middleware

from defi_services.constants.query_constant import Query
from defi_services.constants.token_constant import Token, ProtocolNFT
from defi_services.utils.thread_proxy import ThreadLocalProxy, get_provider_from_uri

logger = logging.getLogger("StateQuerier")


class StateQuerier:
    def __init__(self, provider_uri):
        self.provider_uri = provider_uri
        self._w3 = Web3(Web3.HTTPProvider(provider_uri))
        self._w3.middleware_onion.inject(geth_poa_middleware, layer=0)
        self.client_querier = ClientQuerier(provider_url=provider_uri)

    def get_w3(self):
        return self._w3

    def get_client_querier(self):
        return self.client_querier

    @staticmethod
    def get_function_info(address: str, abi: list, fn_name: str, fn_paras: list = None, block_number: int = 'latest'):
        if fn_paras is None:
            fn_paras = []
        data = {
            "address": address,
            "abi": abi,
            "function": fn_name,
            "params": fn_paras,
            "block_number": block_number
        }

        return data

    @staticmethod
    def get_native_token_balance_info(fn_paras: str = None, block_number: int = 'latest'):
        data = {
            "params": fn_paras,
            "block_number": block_number
        }
        return data

    def query_state_data(self, queries: dict, batch_size: int = 100, workers: int = 5, ignore_error: bool = False):
        """
        Args:
            queries: dict - defi state queries
                - key: str - id of query
                - value: dict - input of query
                    {
                        address: str - address of contract
                        abi: list - abi of contract,
                        function: str - name of the function,
                        params: list - list parameters of function,
                        block_number: int - the block number saving the state data
                    }
            batch_size: int - number of query in each batch queries
            workers: int - maximum number of vCPU used in queries
            ignore_error: bool - ignore error when decode result or not

        Return:
            + A dictionary result of queries
                - key: str - id of query
                - value: result of query
        """
        list_rpc_call, list_call_id = [], []
        for key, value in queries.items():
            fn_paras = value.get(Query.params)
            block_number = value.get(Query.block_number)
            items = key.split('_')
            if Token.native_token == items[2] and "balanceof" == items[0]:
                eth_call = self.add_native_token_balance_rpc_call(fn_paras, key, block_number)
            else:
                abi = value.get(Query.abi)
                contract_address = value.get(Query.address)
                fn_name = value.get(Query.function)
                if not block_number or block_number == 'latest':
                    eth_call = self.add_rpc_call(
                        abi=abi, fn_name=fn_name, contract_address=contract_address,
                        fn_paras=fn_paras, call_id=key
                    )
                else:
                    eth_call = self.add_rpc_call(
                        abi=abi, fn_name=fn_name, contract_address=contract_address,
                        block_number=block_number, fn_paras=fn_paras, call_id=key
                    )
            list_call_id.append(key)
            list_rpc_call.append(eth_call)

        response_data = self.client_querier.sent_batch_to_provider(list_rpc_call, batch_size, workers)
        filtered_response_data = {}
        # loại bỏ những phần tử không có data
        for key, value in response_data.items():
            if value is not None:
                filtered_response_data[key] = value
            else:
                print(key)
        filtered_keys = list(filtered_response_data.keys())
        response_data = filtered_response_data
        list_call_id = [call_id for call_id in list_call_id if call_id in filtered_keys]
        decoded_data = self.decode_response_data(response_data, list_call_id, ignore_error=ignore_error)
        return decoded_data

    @staticmethod
    def add_native_token_balance_rpc_call(
            fn_paras: str = None, call_id: str = None, block_number: int = "latest"):
        eth_call = GetBalance(Web3.to_checksum_address(fn_paras), block_number, call_id)
        return eth_call

    def add_rpc_call(self, abi: dict, fn_name: str, contract_address: str,
                     fn_paras: list = None, call_id: str = None, block_number: int = "latest"):
        args = []
        if fn_paras:
            for item in fn_paras:
                if self._w3.is_address(item):
                    item = self._w3.to_checksum_address(item)
                args.append(item)

        data_call = encode_eth_call_data(abi=abi, fn_name=fn_name, args=args)
        eth_call = EthCall(to=self._w3.to_checksum_address(contract_address), block_number=block_number,
                           data=data_call, abi=abi, fn_name=fn_name, id=call_id)
        return eth_call

    def decode_response_data(self, response_data: dict, list_call_id: list, ignore_error=False):
        decoded_data = {}
        for call_id in list_call_id:
            try:
                response_datum = response_data.get(call_id)
                decoded_datum = response_datum.decode_result()
            except Exception as e:
                decoded_datum = self.check_data(call_id, response_datum)
                if decoded_datum is not None:
                    decoded_data[call_id] = decoded_datum
                    continue
                if not ignore_error:
                    logger.error(f"An exception when decode data from provider: {e}")
                    raise
                else:
                    logger.error(f"[Ignored] An exception when decode data from provider: {e}")
                    continue
            if isinstance(decoded_datum, int):
                decoded_data[call_id] = decoded_datum
            elif len(decoded_datum) == 1:
                decoded_data[call_id] = decoded_datum[0]
            else:
                decoded_data[call_id] = decoded_datum
        return decoded_data

    def create_batch_w3_provider(self):
        return ThreadLocalProxy(lambda: get_provider_from_uri(self.provider_uri, batch=True))

    @staticmethod
    def generate_json_rpc(method, params, request_id=1):
        return {
            'jsonrpc': '2.0',
            'method': method,
            'params': params,
            'id': request_id,
        }

    @staticmethod
    def check_data(call_id: str, data):
        keys = call_id.split("_")
        fn = keys[0]
        if fn == "underlying" and data.result == "0x":
            return Token.native_token
        if fn == "decimals" and keys[1] in ProtocolNFT.nft:
            return 0
        return None



import logging
from logging.handlers import RotatingFileHandler

FORMATTER = logging.Formatter(fmt='[%(asctime)s] [%(levelname)s] [%(name)s] - %(message)s',
                              datefmt='%m-%d-%Y %H:%M:%S %Z')
LOG_FILE = 'logging.log'


def get_console_handler():
    console_handler = logging.StreamHandler()
    console_handler.setFormatter(FORMATTER)
    return console_handler


def get_file_handler():
    file_handler = RotatingFileHandler(LOG_FILE, maxBytes=2000, backupCount=1)
    file_handler.setFormatter(FORMATTER)
    return file_handler


def get_logger(logger_name):
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.DEBUG)
    logger.addHandler(get_console_handler())
    # logger.addHandler(get_file_handler())
    return logger
